'''
Created on Nov 26, 2020

@author: roman.schroeder
'''

# from __future__ import absolute_import
# 
# import celery
# import redis
# 
# # from .worker import app
# from .worker import addtask
# 
# # @app.task
# 
# 
# #@celery.shared_task
# # def add(x, y):
# #     return x + y
# 
# 
# class CustomTask(celery.Task):
#     def run(self, x, y):
#         return x + y
# 
# 
# 
# class MyTasks(object):
#     def add(self, x, y):
#         return x + y
#     def add_one(self, x):
#         return x + 1
#     def add_two(self, x):
#         return x + 2
# 
# 
# def main():
#     pool = redis.ConnectionPool(host='localhost', port=6379, db=0)
#     r = redis.Redis(connection_pool=pool)
#     print(r.keys())
#     #add = CustomTask()
#     #
#     result = addtask.delay(1, 3)
#     print()
#     print(result.ready())
#     print(result.result)
#     print(result.task_id)
#     print(result.get())
#     #
#     print(result.forget())
#     print(r.keys())
# 
# 
# # if __name__ == '__main__':
# #     main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.__init__
'''
:author:    madkote
:contact:   madkote(at)bluewin.ch
:copyright: Copyright 2023, madkote

fastapi_plugins
---------------
FastAPI plugins
'''

from __future__ import absolute_import

from .plugin import *           # noqa F401 F403
from .control import *          # noqa F401 F403
from .logger import *           # noqa F401 F403
from .middleware import *       # noqa F401 F403
from ._redis import *           # noqa F401 F403
from .scheduler import *        # noqa F401 F403
from .settings import *         # noqa F401 F403
from .version import VERSION

# try:
#     import aioredis  # noqa F401
# except ImportError:
#     pass
# else:
#     from ._redis import *  # noqa F401 F403
#
# try:
#     import aiojobs  # noqa F401
# except ImportError:
#     pass
# else:
#     from .scheduler import *  # noqa F401 F403

__author__ = 'madkote <madkote(at)bluewin.ch>'
__version__ = '.'.join(str(x) for x in VERSION)
__copyright__ = 'Copyright 2023, madkote'

# TODO: provide a generic cache type (redis, memcached, in-memory)
#       and share some settings. Module/Sub-Pack cache

# TODO: databases

# TODO: mq - activemq, rabbitmq, kafka
#   -> publish(topic, message, headers)
#   -> consume(topic, callback)

# TODO: celery

# TODO: look at fastapi-cache (memcache?) look at mqtt?

# ... more?



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_memcached

from __future__ import absolute_import

import uuid

import fastapi
import fastapi_plugins
import pytest

from fastapi_plugins.memcached import memcached_plugin, MemcachedSettings


pytestmark = [pytest.mark.anyio, pytest.mark.memcached]


@pytest.fixture
def memcache_config():
    return MemcachedSettings(
        memcached_prestart_tries=10,
        memcached_prestart_wait=1
    )


@pytest.fixture
async def memcache_app(memcache_config):
    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    await memcached_plugin.init_app(app=app, config=memcache_config)
    await memcached_plugin.init()
    yield app
    await memcached_plugin.terminate()


async def test_connect(memcache_app):
    pass


async def test_ping(memcache_app):
    assert (await (await memcached_plugin()).ping()) is not None


async def test_health(memcache_app):
    assert await memcached_plugin.health() == dict(
        host=memcached_plugin.config.memcached_host,
        port=memcached_plugin.config.memcached_port,
        version='1.6.18'
    )


async def test_version(memcache_app):
    assert await (await memcached_plugin()).version() is not None


async def test_get_set(memcache_app):
    c = await memcached_plugin()
    value = str(uuid.uuid4()).encode()
    assert await c.set(b'x', value) is not None
    assert await c.get(b'x') == value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.settings

from __future__ import absolute_import

import functools
import typing

import fastapi
import pydantic
import starlette.config

from .plugin import PluginError
from .plugin import Plugin
from .utils import Annotated
from .version import VERSION

__all__ = [
    'ConfigError', 'ConfigPlugin', 'depends_config', 'config_plugin',
    'TConfigPlugin',
    #
    'register_config',
    'register_config_docker', 'register_config_local', 'register_config_test',
    # 'register_config_by_name',
    'reset_config', 'get_config',
    #
    'registered_configuration', 'registered_configuration_docker',
    'registered_configuration_local', 'registered_configuration_test',
    # 'registered_configuration_by_name',
    #
    'DEFAULT_CONFIG_ENVVAR', 'DEFAULT_CONFIG_NAME',
    'CONFIG_NAME_DEFAULT', 'CONFIG_NAME_DOCKER', 'CONFIG_NAME_LOCAL',
    'CONFIG_NAME_TEST',
]
__author__ = 'madkote <madkote(at)bluewin.ch>'
__version__ = '.'.join(str(x) for x in VERSION)
__copyright__ = 'Copyright 2021, madkote RES'

DEFAULT_CONFIG_ENVVAR: str = 'CONFIG_NAME'
DEFAULT_CONFIG_NAME: str = 'docker'

CONFIG_NAME_DEFAULT = DEFAULT_CONFIG_NAME
CONFIG_NAME_DOCKER = DEFAULT_CONFIG_NAME
CONFIG_NAME_LOCAL = 'local'
CONFIG_NAME_TEST = 'test'


class ConfigError(PluginError):
    pass


class ConfigManager(object):
    def __init__(self):
        self._settings_map = {}

    def register(self, name: str, config: pydantic.BaseSettings) -> None:
        self._settings_map[name] = config

    def reset(self) -> None:
        self._settings_map.clear()

    def get_config(
            self,
            config_or_name: typing.Union[str, pydantic.BaseSettings]=None,
            config_name_default: str=DEFAULT_CONFIG_NAME,
            config_name_envvar: str=DEFAULT_CONFIG_ENVVAR
    ) -> pydantic.BaseSettings:
        if isinstance(config_or_name, pydantic.BaseSettings):
            return config_or_name
        if not config_or_name:
            config_or_name = config_name_default
        base_cfg = starlette.config.Config()
        config_or_name = base_cfg(
            config_name_envvar,
            cast=str,
            default=config_or_name
        )
        if config_or_name not in self._settings_map:
            raise ConfigError('Unknown configuration "%s"' % config_or_name)
        return self._settings_map[config_or_name]()


_manager = ConfigManager()


def register_config(config: pydantic.BaseSettings, name: str=None) -> None:
    if not name:
        name = CONFIG_NAME_DEFAULT
    _manager.register(name, config)


def register_config_docker(config: pydantic.BaseSettings) -> None:
    _manager.register(CONFIG_NAME_DOCKER, config)


def register_config_local(config: pydantic.BaseSettings) -> None:
    _manager.register(CONFIG_NAME_LOCAL, config)


def register_config_test(config: pydantic.BaseSettings) -> None:
    _manager.register(CONFIG_NAME_TEST, config)


# def registered_configuration(cls=None, /, *, name: str=None):
def registered_configuration(cls=None, *, name: str=None):
    if not name:
        name = CONFIG_NAME_DEFAULT

    def wrap(kls):
        _manager.register(name, kls)
        return kls

    if cls is None:
        return wrap
    return wrap(cls)


def registered_configuration_docker(cls=None):
    def wrap(kls):
        _manager.register(CONFIG_NAME_DOCKER, kls)
        return kls
    if cls is None:
        return wrap
    return wrap(cls)


def registered_configuration_local(cls=None):
    def wrap(kls):
        _manager.register(CONFIG_NAME_LOCAL, kls)
        return kls
    if cls is None:
        return wrap
    return wrap(cls)


def registered_configuration_test(cls=None):
    def wrap(kls):
        _manager.register(CONFIG_NAME_TEST, kls)
        return kls
    if cls is None:
        return wrap
    return wrap(cls)


def reset_config() -> None:
    _manager.reset()


@functools.lru_cache()
def get_config(
        config_or_name: typing.Union[str, pydantic.BaseSettings]=None,
        config_name_default: str=DEFAULT_CONFIG_NAME,
        config_name_envvar: str=DEFAULT_CONFIG_ENVVAR
) -> pydantic.BaseSettings:
    return _manager.get_config(
        config_or_name=config_or_name,
        config_name_default=config_name_default,
        config_name_envvar=config_name_envvar
    )


class ConfigPlugin(Plugin):
    DEFAULT_CONFIG_CLASS = pydantic.BaseSettings

    async def _on_call(self) -> pydantic.BaseSettings:
        return self.config

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        app.state.PLUGIN_CONFIG = self


config_plugin = ConfigPlugin()


async def depends_config(
    conn: starlette.requests.HTTPConnection
) -> pydantic.BaseSettings:
    return await conn.app.state.PLUGIN_CONFIG()


TConfigPlugin = Annotated[pydantic.BaseSettings, fastapi.Depends(depends_config)]   # noqa E501



'''
#
# normal redis
#
#redis://:password@hostname:port/db_number
app.conf.broker_url = 'redis://localhost:6379/0'

#
# sentinel
#
app.conf.broker_url = 'sentinel://localhost:26379;sentinel://localhost:26380;sentinel://localhost:26381'
app.conf.broker_transport_options = { 'master_name': "cluster1" }


#
# when worker accepts a task
#
app.conf.broker_transport_options = {'visibility_timeout': 3600}  # 1 hour



#
# results
#
app.conf.result_backend = 'redis://localhost:6379/0'
app.conf.result_backend_transport_options = {'master_name': "mymaster"}    # sentinel
'''


'''
Created on Nov 26, 2020

@author: roman.schroeder
'''

from __future__ import absolute_import

import inspect

import celery
import redis

# app = celery.Celery(
#     __name__,
#     broker='redis://localhost:6379/0',
#     backend='redis://localhost:6379/0',
#     # include=['mytasks.tasks']
# )

# from .tasks import CustomTask
# 
# addtask = app.register_task(CustomTask)
# 
# print(addtask)


class MulTasks(object):
    def add(self, x, y):
        return x + y
    def add_one(self, x):
        return x + 1
    def add_two(self, x):
        return x + 2
    def mul(self, x, y):
        return x * y


def make_app():
    return celery.Celery(
        __name__,
        broker='redis://localhost:6379/0',
        backend='redis://localhost:6379/0',
        # include=['mytasks.tasks']
    )


def get_app(*tasks, bind=True):
    app = celery.Celery(
        __name__,
        broker='redis://localhost:6379/0',
        backend='redis://localhost:6379/0',
        # include=['mytasks.tasks']
    )
    for task in tasks:
        print(task)
        for name, m in inspect.getmembers(task, lambda x: inspect.isfunction(x) or inspect.ismethod(x)):
            print('>', name, m)
            print(dir(m))
            print(m.__annotations__)
            setattr(task, name, app.task(m, bind=bind))
            print(getattr(task, name))

#     class _MultiplyTasks(object):
#         @app.task(bind=True)
#         def mul(self, x, y):
#             return x * y
    # raise
    return app


the_app_not_to_use = get_app(MulTasks)


# @app.task
# def add(x, y):
#     return x + y


# class MyTasks(object):
# #     @app.task
# #     def add(self, x, y):
# #         return x + y
#     @app.task(bind=True)
#     def add_one(self, x):
#         return x + 1
#     @app.task(bind=True)
#     def add_two(self, x):
#         return x + 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_control

from __future__ import absolute_import

import contextlib
import typing

import fastapi
import pydantic
import pytest
import starlette.testclient

import fastapi_plugins

from fastapi_plugins.control import DEFAULT_CONTROL_VERSION
from fastapi_plugins.memcached import memcached_plugin
from fastapi_plugins.memcached import MemcachedSettings


pytestmark = [pytest.mark.anyio, pytest.mark.control]


class DummyPluginHealthOK(
        fastapi_plugins.Plugin,
        fastapi_plugins.ControlHealthMixin
):
    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,     # @UnusedVariable
            *args,                                  # @UnusedVariable
            **kwargs                                # @UnusedVariable
    ) -> None:
        app.state.DUMMY_PLUGIN_HEALTH_OK = self

    async def health(self) -> typing.Dict:
        return dict(dummy='OK')


class DummyPluginHealthOKOnce(
        fastapi_plugins.Plugin,
        fastapi_plugins.ControlHealthMixin
):
    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,     # @UnusedVariable
            *args,                                  # @UnusedVariable
            **kwargs                                # @UnusedVariable
    ) -> None:
        self.counter = 0
        app.state.DUMMY_PLUGIN_HEALTH_OK_ONCE = self

    async def health(self) -> typing.Dict:
        if self.counter > 0:
            raise Exception('Health check failed')
        else:
            self.counter += 1
            return dict(dummy='OK')


class DummyPluginHealthFail(
        fastapi_plugins.Plugin,
        fastapi_plugins.ControlHealthMixin
):
    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,     # @UnusedVariable
            *args,                                  # @UnusedVariable
            **kwargs                                # @UnusedVariable
    ) -> None:
        app.state.DUMMY_PLUGIN_HEALTH_FAIL = self

    async def health(self) -> typing.Dict:
        raise Exception('Health check failed')


class DummyPluginHealthNotDefined(
        fastapi_plugins.Plugin,
        fastapi_plugins.ControlHealthMixin
):
    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,     # @UnusedVariable
            *args,                                  # @UnusedVariable
            **kwargs                                # @UnusedVariable
    ) -> None:
        app.state.DUMMY_PLUGIN_NOT_DEFINED = self


class MyControlConfig(
        fastapi_plugins.RedisSettings,
        fastapi_plugins.SchedulerSettings,
        fastapi_plugins.ControlSettings,
        MemcachedSettings
):
    pass


def make_app(config=None, version=None, environ=None, plugins=None):
    if plugins is None:
        plugins = []
    if config is None:
        config = fastapi_plugins.ControlSettings()

    @contextlib.asynccontextmanager
    async def lifespan(app: fastapi.FastAPI):
        for p in plugins:
            await p.init_app(app, config)
            await p.init()
        kwargs = {}
        if version:
            kwargs.update(**dict(version=version))
        if environ:
            kwargs.update(**dict(environ=environ))
        await fastapi_plugins.control_plugin.init_app(app, config, **kwargs)
        await fastapi_plugins.control_plugin.init()
        yield
        await fastapi_plugins.control_plugin.terminate()
        for p in plugins:
            await p.terminate()

    return fastapi_plugins.register_middleware(
        fastapi.FastAPI(lifespan=lifespan)
    )


@pytest.fixture(params=[{}])
def client(request):
    with starlette.testclient.TestClient(make_app(**request.param)) as c:
        yield c


@pytest.mark.parametrize(
    'kwargs, result',
    [
        pytest.param({}, {}),
        pytest.param(dict(environ=dict(ping='pong')), dict(ping='pong')),
    ]
)
async def test_controller_environ(kwargs, result):
    assert result == await fastapi_plugins.Controller(**kwargs).get_environ()


async def test_controller_health():
    assert dict(status=True, checks=[]) == (await fastapi_plugins.Controller().get_health()).dict() # noqa E501


@pytest.mark.parametrize(
    'klass, name, status, details',
    [
        pytest.param(
            DummyPluginHealthOK,
            'DUMMY_PLUGIN_OK',
            True,
            dict(dummy='OK')
        ),
        pytest.param(
            DummyPluginHealthNotDefined,
            'DUMMY_PLUGIN_NOT_DEFINED',
            True,
            {}
        ),
        pytest.param(
            DummyPluginHealthFail,
            'DUMMY_PLUGIN_HEALTH_FAIL',
            False,
            dict(error='Health check failed')
        ),
    ]
)
async def test_controller_health_plugin(
    klass, name, status, details
):
    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    config = fastapi_plugins.ControlSettings()
    dummy = klass()
    await dummy.init_app(app, config)
    await dummy.init()
    try:
        c = fastapi_plugins.Controller()
        c.plugins.append((name, dummy))
        assert (await c.get_health()).dict() == dict(
            status=status,
            checks=[
                dict(
                    name=name,
                    status=status,
                    details=details
                )
            ]
        )
    finally:
        await dummy.terminate()


async def test_controller_heartbeat():
    assert await fastapi_plugins.Controller().get_heart_beat() is True


@pytest.mark.parametrize(
    'kwargs, version',
    [
        pytest.param({}, DEFAULT_CONTROL_VERSION),
        pytest.param(dict(version='1.2.3'), '1.2.3'),
    ]
)
async def test_controller_version(kwargs, version):
    assert version == await fastapi_plugins.Controller(**kwargs).get_version()


@pytest.mark.parametrize(
    'client, result, status, endpoint',
    [
        pytest.param(
            {},
            dict(environ={}),
            200,
            '/control/environ'
        ),
        pytest.param(
            dict(environ=dict(ping='pong')),
            dict(environ=dict(ping='pong')),
            200,
            '/control/environ'
        ),
        pytest.param(
            {},
            {'version': DEFAULT_CONTROL_VERSION},
            200,
            '/control/version'
        ),
        pytest.param(
            dict(version='1.2.3'),
            {'version': '1.2.3'},
            200,
            '/control/version'
        ),
        pytest.param(
            {},
            dict(is_alive=True),
            200,
            '/control/heartbeat'
        ),
        pytest.param(
            {},
            dict(status=True, checks=[]),
            200,
            '/control/health'
        ),
        pytest.param(
            dict(
                plugins=[
                    DummyPluginHealthNotDefined(),
                    DummyPluginHealthOK()
                ]
            ),
            dict(
                status=True,
                checks=[
                    dict(
                        name='DUMMY_PLUGIN_NOT_DEFINED',
                        status=True,
                        details={}
                    ),
                    dict(
                        name='DUMMY_PLUGIN_HEALTH_OK',
                        status=True,
                        details=dict(dummy='OK')
                    )
                ]
            ),
            200,
            '/control/health'
        ),
        pytest.param(
            dict(
                config=MyControlConfig(),
                plugins=[
                    fastapi_plugins.redis_plugin,
                    fastapi_plugins.scheduler_plugin,
                    memcached_plugin
                ]
            ),
            dict(
                status=True,
                checks=[
                    {
                        'name': 'REDIS',
                        'status': True,
                        'details': {
                            'redis_type': 'redis',
                            'redis_address': 'redis://localhost:6379',
                            'redis_pong': True
                        }
                    },
                    {
                        'name': 'AIOJOBS_SCHEDULER',
                        'status': True,
                        'details': {
                            'jobs': 0,
                            'active': 0,
                            'pending': 0,
                            'limit': 100,
                            'closed': False
                        }
                    },
                    {
                        'name': 'MEMCACHED',
                        'status': True,
                        'details': {
                            'host': 'localhost',
                            'port': 11211,
                            'version': '1.6.18'
                        }
                    }
                ]
            ),
            200,
            '/control/health',
        ),
        pytest.param(
            dict(
                plugins=[
                    DummyPluginHealthNotDefined(),
                    DummyPluginHealthOK(),
                    DummyPluginHealthOKOnce()
                ]
            ),
            {
                'detail': {
                    'status': False,
                    'checks': [
                        {
                            'name': 'DUMMY_PLUGIN_NOT_DEFINED',
                            'status': True,
                            'details': {}
                        },
                        {
                            'name': 'DUMMY_PLUGIN_HEALTH_OK',
                            'status': True,
                            'details': {'dummy': 'OK'}
                        },
                        {
                            'name': 'DUMMY_PLUGIN_HEALTH_OK_ONCE',
                            'status': False,
                            'details': {'error': 'Health check failed'}
                        }
                    ]
                }
            },
            417,
            '/control/health'
        ),
    ],
    indirect=['client']
)
async def test_router(client, result, status, endpoint):
    response = client.get(endpoint)
    assert status == response.status_code
    assert result == response.json()


async def test_router_health_with_plugins_broken_init():
    with pytest.raises(fastapi_plugins.ControlError):
        with starlette.testclient.TestClient(
            make_app(
                plugins=[
                    DummyPluginHealthFail(),
                    DummyPluginHealthOK()
                ]
            )
        ) as c:
            c.get('/control/version')


@pytest.mark.parametrize(
    'client, result, status, endpoint',
    [
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_router_prefix='outofcontrol'
                )
            ),
            dict(status=True, checks=[]),
            200,
            '/outofcontrol/health'
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_router_prefix='outofcontrol'
                )
            ),
            {'version': DEFAULT_CONTROL_VERSION},
            200,
            '/outofcontrol/version'
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_router_prefix='outofcontrol'
                ),
                version='1.2.3'
            ),
            {'version': '1.2.3'},
            200,
            '/outofcontrol/version'
        ),
    ],
    indirect=['client']
)
async def test_router_prefix_custom(client, result, status, endpoint):
    response = client.get(endpoint)
    assert status == response.status_code
    assert result == response.json()


@pytest.mark.parametrize(
    'client, endpoints',
    [
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_enable_environ=False,
                    control_enable_health=False,
                    control_enable_heartbeat=False,
                    control_enable_version=False
                )
            ),
            [
                (404, '/control/environ'),
                (404, '/control/version'),
                (404, '/control/heartbeat'),
                (404, '/control/health')
            ]
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_enable_environ=False,
                    control_enable_health=True,
                    control_enable_heartbeat=True,
                    control_enable_version=True
                )
            ),
            [
                (404, '/control/environ'),
                (200, '/control/version'),
                (200, '/control/heartbeat'),
                (200, '/control/health')
            ]
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_enable_environ=True,
                    control_enable_health=True,
                    control_enable_heartbeat=True,
                    control_enable_version=False
                )
            ),
            [
                (200, '/control/environ'),
                (404, '/control/version'),
                (200, '/control/heartbeat'),
                (200, '/control/health')
            ]
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_enable_environ=True,
                    control_enable_health=False,
                    control_enable_heartbeat=True,
                    control_enable_version=True
                )
            ),
            [
                (200, '/control/environ'),
                (200, '/control/version'),
                (200, '/control/heartbeat'),
                (404, '/control/health')
            ]
        ),
        pytest.param(
            dict(
                config=fastapi_plugins.ControlSettings(
                    control_enable_environ=True,
                    control_enable_health=True,
                    control_enable_heartbeat=False,
                    control_enable_version=True
                )
            ),
            [
                (200, '/control/environ'),
                (200, '/control/version'),
                (404, '/control/heartbeat'),
                (200, '/control/health')
            ]
        ),
    ],
    indirect=['client']
)
async def test_router_disable(client, endpoints):
    for status, endpoint in endpoints:
        response = client.get(endpoint)
        assert status == response.status_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_redis

from __future__ import absolute_import

import uuid

import fastapi
import pytest

import fastapi_plugins
from fastapi_plugins._redis import RedisType


pytestmark = [pytest.mark.anyio, pytest.mark.redis]


@pytest.fixture(params=None)
async def redisapp(request):
    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    await fastapi_plugins.redis_plugin.init_app(
        app=app,
        config=request.param or fastapi_plugins.RedisSettings(),
    )
    await fastapi_plugins.redis_plugin.init()
    yield app
    await fastapi_plugins.redis_plugin.terminate()


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(
            fastapi_plugins.RedisSettings(redis_url='redis://localhost:6379/1')
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='fakeredis',
                redis_url='redis://localhost:6379/1'
            ),
            marks=pytest.mark.fakeredis
        ),
    ],
    indirect=['redisapp']
)
async def test_connect_redis_url(redisapp):
    pass


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(fastapi_plugins.RedisSettings()),
        pytest.param(
            fastapi_plugins.RedisSettings(redis_type='fakeredis'),
            marks=pytest.mark.fakeredis
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='sentinel',
                redis_sentinels='localhost:26379'
            ),
            marks=pytest.mark.sentinel
        ),
    ],
    indirect=['redisapp']
)
async def test_connect(redisapp):
    pass


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(fastapi_plugins.RedisSettings()),
        pytest.param(
            fastapi_plugins.RedisSettings(redis_type='fakeredis'),
            marks=pytest.mark.fakeredis
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='sentinel',
                redis_sentinels='localhost:26379'
            ),
            marks=pytest.mark.sentinel
        ),
    ],
    indirect=['redisapp']
)
async def test_ping(redisapp):
    assert (await (await fastapi_plugins.redis_plugin()).ping()) is True


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(fastapi_plugins.RedisSettings()),
        pytest.param(
            fastapi_plugins.RedisSettings(redis_type='fakeredis'),
            marks=pytest.mark.fakeredis
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='sentinel',
                redis_sentinels='localhost:26379'
            ),
            marks=pytest.mark.sentinel
        ),
    ],
    indirect=['redisapp']
)
async def test_health(redisapp):
    assert await fastapi_plugins.redis_plugin.health() == dict(
        redis_type=fastapi_plugins.redis_plugin.config.redis_type,
        redis_address=fastapi_plugins.redis_plugin.config.get_sentinels() if fastapi_plugins.redis_plugin.config.redis_type == RedisType.sentinel else fastapi_plugins.redis_plugin.config.get_redis_address(), # noqa E501
        redis_pong=True
    )


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(fastapi_plugins.RedisSettings()),
        pytest.param(
            fastapi_plugins.RedisSettings(redis_type='fakeredis'),
            marks=pytest.mark.fakeredis
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='sentinel',
                redis_sentinels='localhost:26379'
            ),
            marks=pytest.mark.sentinel
        ),
    ],
    indirect=['redisapp']
)
async def test_get_set(redisapp):
    c = await fastapi_plugins.redis_plugin()
    value = str(uuid.uuid4())
    assert await c.set('x', value) is not None
    assert await c.get('x') == value


@pytest.mark.parametrize(
    'redisapp',
    [
        pytest.param(fastapi_plugins.RedisSettings(redis_ttl=61)),
        pytest.param(
            fastapi_plugins.RedisSettings(redis_type='fakeredis'),
            marks=pytest.mark.fakeredis
        ),
        pytest.param(
            fastapi_plugins.RedisSettings(
                redis_type='sentinel',
                redis_sentinels='localhost:26379'
            ),
            marks=pytest.mark.sentinel
        ),
    ],
    indirect=['redisapp']
)
async def test_get_set_ttl(redisapp):
    c = await fastapi_plugins.redis_plugin()
    value = str(uuid.uuid4())
    assert await c.setex('x', c.TTL, value) is not None
    assert await c.get('x') == value
    assert await c.ttl('x') == fastapi_plugins.redis_plugin.config.redis_ttl


# def redis_must_be_running(cls):
#     # TODO: This SHOULD be improved
#     try:
#         r = redis.StrictRedis('localhost', port=6379)
#         r.ping()
#     except redis.ConnectionError:
#         redis_running = False
#     else:
#         redis_running = True
#     if not redis_running:
#         for name, attribute in inspect.getmembers(cls):
#             if name.startswith('test_'):
#                 @wraps(attribute)
#                 def skip_test(*args, **kwargs):
#                     pytest.skip("Redis is not running.")
#                 setattr(cls, name, skip_test)
#         cls.setUp = lambda x: None
#         cls.tearDown = lambda x: None
#     return cls



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# myapp.celeryp.app
'''
:author:    roman-telepathy-ai
:contact:   roman.schroeder(at)telepathy.ai
:copyright: Copyright 2020, Telepathy Labs

myapp.celeryp.app
---------
Module
'''

from __future__ import absolute_import

import celery

from .config import BROKER_CONN_URI, BACKEND_CONN_URI

VERSION = (0, 1, 0)

__all__ = []
__author__ = 'roman-telepathy-ai <roman.schroeder(at)telepathy.ai>'
__version__ = '.'.join(str(x) for x in VERSION)
__copyright__ = 'Copyright 2020, Telepathy Labs'


app = celery.Celery(__name__, broker=BROKER_CONN_URI, backend=BACKEND_CONN_URI)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_settings

from __future__ import absolute_import

import os

import fastapi
import pytest

import fastapi_plugins

from fastapi_plugins.settings import ConfigManager


pytestmark = [pytest.mark.anyio, pytest.mark.settings]


@pytest.fixture
def myconfig():
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    return MyConfig


@pytest.fixture
def myconfig_name():
    return 'myconfig'


@pytest.fixture
def myconfig_name_default():
    return fastapi_plugins.CONFIG_NAME_DEFAULT


@pytest.fixture
def plugin():
    fastapi_plugins.reset_config()
    fastapi_plugins.get_config.cache_clear()
    yield fastapi_plugins
    fastapi_plugins.reset_config()
    fastapi_plugins.get_config.cache_clear()


def test_manager_register(plugin, myconfig, myconfig_name):
    m = ConfigManager()
    m.register(myconfig_name, myconfig)
    assert m._settings_map == {myconfig_name: myconfig}


def test_manager_reset(plugin, myconfig, myconfig_name):
    m = ConfigManager()
    m.register(myconfig_name, myconfig)
    assert m._settings_map == {myconfig_name: myconfig}
    m.reset()
    assert m._settings_map == {}


@pytest.mark.parametrize(
    'name',
    [
        pytest.param('myconfig_name'),
        pytest.param('myconfig_name_default'),
    ]
)
def test_manager_get_config(name, plugin, myconfig, request):
    name = request.getfixturevalue(name)
    m = ConfigManager()
    m.register(name, myconfig)
    assert m._settings_map == {name: myconfig}
    assert myconfig().dict() == m.get_config(name).dict()


def test_manager_get_config_not_existing(plugin, myconfig, myconfig_name):
    m = ConfigManager()
    m.register(myconfig_name, myconfig)
    assert m._settings_map == {myconfig_name: myconfig}
    with pytest.raises(fastapi_plugins.ConfigError):
        m.get_config()


def test_wrap_register_config(plugin, myconfig):
    fastapi_plugins.register_config(myconfig)
    assert myconfig().dict() == fastapi_plugins.get_config().dict()


def test_wrap_register_config_docker(plugin, myconfig):
    # docker is default
    fastapi_plugins.register_config_docker(myconfig)
    assert myconfig().dict() == fastapi_plugins.get_config().dict()
    assert myconfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_DOCKER).dict()    # noqa E501


def test_wrap_register_config_local(plugin, myconfig):
    fastapi_plugins.register_config_local(myconfig)
    assert myconfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_LOCAL).dict()    # noqa E501
    os.environ[fastapi_plugins.DEFAULT_CONFIG_ENVVAR] = fastapi_plugins.CONFIG_NAME_LOCAL               # noqa E501
    try:
        assert myconfig().dict() == fastapi_plugins.get_config().dict()
    finally:
        os.environ.pop(fastapi_plugins.DEFAULT_CONFIG_ENVVAR)


def test_wrap_register_config_test(plugin, myconfig):
    fastapi_plugins.register_config_test(myconfig)
    assert myconfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_TEST).dict() # noqa E501
    os.environ[fastapi_plugins.DEFAULT_CONFIG_ENVVAR] = fastapi_plugins.CONFIG_NAME_TEST            # noqa E501
    try:
        assert myconfig().dict() == fastapi_plugins.get_config().dict()
    finally:
        os.environ.pop(fastapi_plugins.DEFAULT_CONFIG_ENVVAR)


def test_wrap_register_config_by_name(plugin, myconfig, myconfig_name):
    fastapi_plugins.register_config(myconfig, name=myconfig_name)
    assert myconfig().dict() == fastapi_plugins.get_config(myconfig_name).dict()    # noqa E501


def test_wrap_get_config(plugin, myconfig):
    fastapi_plugins.register_config(myconfig)
    assert myconfig().dict() == fastapi_plugins.get_config().dict()


def test_wrap_get_config_by_name(plugin, myconfig, myconfig_name):
    fastapi_plugins.register_config(myconfig, name=myconfig_name)
    assert myconfig().dict() == fastapi_plugins.get_config(myconfig_name).dict()    # noqa E501


def test_wrap_reset_config(plugin, myconfig):
    from fastapi_plugins.settings import _manager
    assert _manager._settings_map == {}
    fastapi_plugins.register_config(myconfig)
    assert _manager._settings_map == {fastapi_plugins.CONFIG_NAME_DOCKER: myconfig} # noqa E501
    fastapi_plugins.reset_config()
    assert _manager._settings_map == {}


def test_decorator_register_config(plugin):
    @fastapi_plugins.registered_configuration
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    assert MyConfig().dict() == fastapi_plugins.get_config().dict()


def test_decorator_register_config_docker(plugin):
    @fastapi_plugins.registered_configuration_docker
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    # docker is default
    assert MyConfig().dict() == fastapi_plugins.get_config().dict()
    assert MyConfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_DOCKER).dict()   # noqa E501


def test_decorator_register_config_local(plugin):
    @fastapi_plugins.registered_configuration_local
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    assert MyConfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_LOCAL).dict()    # noqa E501
    os.environ[fastapi_plugins.DEFAULT_CONFIG_ENVVAR] = fastapi_plugins.CONFIG_NAME_LOCAL               # noqa E501
    try:
        assert MyConfig().dict() == fastapi_plugins.get_config().dict()
    finally:
        os.environ.pop(fastapi_plugins.DEFAULT_CONFIG_ENVVAR)


def test_decorator_register_config_test(plugin):
    @fastapi_plugins.registered_configuration_test
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    assert MyConfig().dict() == fastapi_plugins.get_config(fastapi_plugins.CONFIG_NAME_TEST).dict() # noqa E501
    os.environ[fastapi_plugins.DEFAULT_CONFIG_ENVVAR] = fastapi_plugins.CONFIG_NAME_TEST            # noqa E501
    try:
        assert MyConfig().dict() == fastapi_plugins.get_config().dict()
    finally:
        os.environ.pop(fastapi_plugins.DEFAULT_CONFIG_ENVVAR)


def test_decorator_register_config_by_name(plugin, myconfig_name):
    @fastapi_plugins.registered_configuration(name=myconfig_name)
    class MyConfig(fastapi_plugins.PluginSettings):
        api_name: str = 'API name'

    assert MyConfig().dict() == fastapi_plugins.get_config(myconfig_name).dict()    # noqa E501


async def test_app_config(plugin):
    @fastapi_plugins.registered_configuration
    class MyConfigDocker(fastapi_plugins.PluginSettings):
        api_name: str = 'docker'

    @fastapi_plugins.registered_configuration_local
    class MyConfigLocal(fastapi_plugins.PluginSettings):
        api_name: str = 'local'

    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    config = fastapi_plugins.get_config()
    await fastapi_plugins.config_plugin.init_app(app=app, config=config)
    await fastapi_plugins.config_plugin.init()
    try:
        assert MyConfigDocker().dict() == (await fastapi_plugins.config_plugin()).dict()    # noqa E501
    finally:
        await fastapi_plugins.config_plugin.terminate()
        fastapi_plugins.reset_config()


async def test_app_config_environ(plugin):
    os.environ[fastapi_plugins.DEFAULT_CONFIG_ENVVAR] = fastapi_plugins.CONFIG_NAME_LOCAL   # noqa E501
    try:
        @fastapi_plugins.registered_configuration
        class MyConfigDocker(fastapi_plugins.PluginSettings):
            api_name: str = 'docker'

        @fastapi_plugins.registered_configuration_local
        class MyConfigLocal(fastapi_plugins.PluginSettings):
            api_name: str = 'local'

        app = fastapi_plugins.register_middleware(fastapi.FastAPI())
        config = fastapi_plugins.get_config()
        await fastapi_plugins.config_plugin.init_app(app=app, config=config)
        await fastapi_plugins.config_plugin.init()
        try:
            assert MyConfigLocal().dict() == (await fastapi_plugins.config_plugin()).dict() # noqa E501
        finally:
            await fastapi_plugins.config_plugin.terminate()
    finally:
        os.environ.pop(fastapi_plugins.DEFAULT_CONFIG_ENVVAR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_scheduler

from __future__ import absolute_import

import asyncio
import contextlib
import typing
import uuid

import fastapi
import pytest
import starlette.testclient

import fastapi_plugins


pytestmark = [pytest.mark.anyio, pytest.mark.scheduler]


def make_app(config=None):
    if config is None:
        class AppSettings(
                fastapi_plugins.RedisSettings,
                fastapi_plugins.SchedulerSettings
        ):
            api_name: str = str(__name__)
        config = AppSettings()

    @contextlib.asynccontextmanager
    async def lifespan(app: fastapi.FastAPI):
        await fastapi_plugins.redis_plugin.init_app(app, config=config)
        await fastapi_plugins.redis_plugin.init()
        await fastapi_plugins.scheduler_plugin.init_app(app, config)
        await fastapi_plugins.scheduler_plugin.init()
        yield
        await fastapi_plugins.scheduler_plugin.terminate()
        await fastapi_plugins.redis_plugin.terminate()

    app = fastapi_plugins.register_middleware(
        fastapi.FastAPI(lifespan=lifespan)
    )

    @app.post('/jobs/schedule')
    async def job_post(
        cache: fastapi_plugins.TRedisPlugin,
        scheduler: fastapi_plugins.TSchedulerPlugin,
        timeout: int=fastapi.Query(..., title='the job sleep time')
    ) -> str:
        async def coro(job_id, timeout, cache):
            await cache.set(job_id, 'processing')
            try:
                await asyncio.sleep(timeout)
                if timeout >= 3:
                    raise Exception('ugly error')
            except asyncio.CancelledError:
                await cache.set(job_id, 'canceled')
            except Exception:
                await cache.set(job_id, 'erred')
            else:
                await cache.set(job_id, 'success')

        job_id = str(uuid.uuid4()).replace('-', '')
        await cache.set(job_id, 'pending')
        await scheduler.spawn(coro(job_id, timeout, cache))
        return job_id

    @app.get('/jobs/status')
    async def job_get(
        cache: fastapi_plugins.TRedisPlugin,
        job_id: str=fastapi.Query(..., title='the job id')
    ) -> typing.Dict:
        status = await cache.get(job_id)
        if status is None:
            raise fastapi.HTTPException(
                status_code=starlette.status.HTTP_404_NOT_FOUND,
                detail='Job %s not found' % job_id
            )
        return dict(job_id=job_id, status=status)

    @app.get('/demo')
    async def demo_get(
    ) -> typing.Dict:
        return dict(demo=123)

    return app


@pytest.fixture(params=[{}])
def client(request):
    with starlette.testclient.TestClient(make_app(**request.param)) as c:
        yield c


@pytest.fixture
async def schedulerapp():
    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    config = fastapi_plugins.SchedulerSettings()
    await fastapi_plugins.scheduler_plugin.init_app(app=app, config=config)
    await fastapi_plugins.scheduler_plugin.init()
    yield app
    await fastapi_plugins.scheduler_plugin.terminate()


async def test_basic(schedulerapp):
    count = 3
    res = {}

    async def coro(name, timeout):
        try:
            await asyncio.sleep(timeout)
            res[name] = name
        except asyncio.CancelledError as e:
            res[name] = 'cancel'
            raise e

    s = await fastapi_plugins.scheduler_plugin()
    for i in range(count):
        await s.spawn(coro(str(i), i / 10))
    await asyncio.sleep(1)
    assert res == dict([(str(i), str(i)) for i in range(count)])


async def test_health(schedulerapp):
    count = 3
    res = {}

    async def coro(name, timeout):
        try:
            await asyncio.sleep(timeout)
            res[name] = name
        except asyncio.CancelledError as e:
            res[name] = 'cancel'
            raise e

    s = await fastapi_plugins.scheduler_plugin()
    for i in range(count):
        await s.spawn(coro(str(i), i / 10))
    await asyncio.sleep(1)
    assert res == dict([(str(i), str(i)) for i in range(count)])
    assert await fastapi_plugins.scheduler_plugin.health() == dict(
        jobs=0,
        active=0,
        pending=0,
        limit=100,
        closed=False
    )


async def test_endpoints(client):
    endpoint = '/demo'
    response = client.get(endpoint)
    assert 200 == response.status_code
    #
    job_timeout = 1
    print()
    print('> submit a job with timeout=%s' % job_timeout)
    endpoint = '/jobs/schedule?timeout=%s' % job_timeout
    response = client.post(endpoint)
    assert 200 == response.status_code
    job_id = response.json()
    assert len(job_id) > 0
    #
    print('> check the job %s' % job_id)
    endpoint = '/jobs/status?job_id=%s' % job_id
    response = client.get(endpoint)
    assert 200 == response.status_code
    assert response.json() == {'job_id': job_id, 'status': 'processing'}
    #
    tries = 10000
    attempt = 0
    endpoint = '/jobs/status?job_id=%s' % job_id
    while attempt < tries:
        response = client.get(endpoint)
        assert 200 == response.status_code
        if response.json() == {'job_id': job_id, 'status': 'success'}:
            break
        attempt += 1
    else:
        pytest.fail(f'job {job_id} with timeout {job_timeout} not finished')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.logger

from __future__ import absolute_import

import datetime
import enum
import logging
import numbers
import queue
import sys
import typing

import fastapi
import pydantic
import starlette.requests

from pythonjsonlogger import jsonlogger

from .plugin import PluginError
from .plugin import PluginSettings
from .plugin import Plugin

from .control import ControlHealthMixin
from .utils import Annotated

__all__ = [
    'LoggingError', 'LoggingStyle', 'LoggingHandlerType', 'LoggingSettings',
    'LoggingPlugin', 'log_plugin', 'log_adapter', 'depends_logging',
    'TLoggerPlugin'
]


class QueueHandler(logging.Handler):
    def __init__(self, *args, mqueue=None, **kwargs):
        super(QueueHandler, self).__init__(*args, **kwargs)
        self.mqueue = mqueue if mqueue is not None else queue.Queue()

    def emit(self, record):
        self.mqueue.put(self.format(record))


class JsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super(JsonFormatter, self).add_fields(
            log_record,
            record,
            message_dict
        )
        if not log_record.get('timestamp'):
            # this doesn't use record.created, so it is slightly off
            now = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['timestamp'] = now
        if log_record.get('level'):
            log_record['level'] = log_record['level'].upper()
        else:
            log_record['level'] = record.levelname
        if not log_record.get('name'):
            log_record['name'] = record.name


class LogfmtFormatter(logging.Formatter):
    # from https://github.com/jkakar/logfmt-python
    def format_line(self, extra: typing.Dict) -> str:
        outarr = []
        for k, v in extra.items():
            if v is None:
                outarr.append('%s=' % k)
                continue
            if isinstance(v, bool):
                v = 'true' if v else 'false'
            elif isinstance(v, numbers.Number):
                pass
            else:
                if isinstance(v, (dict, object)):
                    v = str(v)
                v = '"%s"' % v.replace('"', '\\"')
            outarr.append('%s=%s' % (k, v))
        return ' '.join(outarr)

    def format(self, record):
        return ' '.join(
            [
                'at=%s' % record.levelname,
                'msg="%s"' % record.getMessage().replace('"', '\\"'),
                'process=%s' % record.processName,
                self.format_line(getattr(record, 'context', {})),
            ]
        ).strip()


class LoggerLogfmt(logging.Logger):
    def makeRecord(
            self,
            name,
            level,
            fn,
            lno,
            msg,
            args,
            exc_info,
            func=None,
            extra=None,
            sinfo=None
    ):
        factory = logging.getLogRecordFactory()
        rv = factory(name, level, fn, lno, msg, args, exc_info, func, sinfo)
        if extra is not None:
            rv.__dict__['context'] = dict(**extra)
        return rv


class LoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        _extra = kwargs.get('extra', {})
        kwargs["extra"] = dict(**self.extra)
        kwargs["extra"].update(**_extra)
        return msg, kwargs


class LoggingError(PluginError):
    pass


@enum.unique
class LoggingStyle(str, enum.Enum):
    logfmt = 'logfmt'
    logjson = 'json'
    logtxt = 'txt'


@enum.unique
class LoggingHandlerType(str, enum.Enum):
    loglist = 'list'
    logstdout = 'stdout'


class LoggingSettings(PluginSettings):
    logging_level: int = logging.WARNING
    logging_style: LoggingStyle = LoggingStyle.logtxt
    logging_handler: LoggingHandlerType = LoggingHandlerType.logstdout
    logging_fmt: str = None


class LoggingPlugin(Plugin, ControlHealthMixin):
    DEFAULT_CONFIG_CLASS: pydantic.BaseSettings = LoggingSettings

    def _create_logger(
            self,
            name: str,
            config: pydantic.BaseSettings=None
    ) -> logging.Logger:
        logger_klass = None
        #
        if config.logging_handler == LoggingHandlerType.logstdout:
            handler = logging.StreamHandler(stream=sys.stdout)
        elif config.logging_handler == LoggingHandlerType.loglist:
            handler = QueueHandler()
        #
        if config.logging_style == LoggingStyle.logtxt:
            formatter = logging.Formatter(fmt=config.logging_fmt)
        elif config.logging_style == LoggingStyle.logfmt:
            formatter = LogfmtFormatter()
            logger_klass = LoggerLogfmt
        elif config.logging_style == LoggingStyle.logjson:
            formatter = JsonFormatter()
        else:
            raise LoggingError(
                'unknown logging format style %s' % config.logging_style
            )
        if logger_klass is not None:
            _original_logger_klass = logging.getLoggerClass()
            try:
                logging.setLoggerClass(logger_klass)
                logger = logging.getLogger(name)
            finally:
                logging.setLoggerClass(_original_logger_klass)
        else:
            logger = logging.getLogger(name)
        #
        logger.setLevel(config.logging_level)
        handler.setLevel(config.logging_level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    def _on_init(self) -> None:
        self.config = None
        self.logger = None

    async def _on_call(self) -> logging.Logger:
        if self.logger is None:
            raise LoggingError('Logging is not initialized')
        return self.logger

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,
            *,
            name: str=None
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        if self.config is None:
            raise LoggingError('Logging configuration is not initialized')
        elif not isinstance(self.config, self.DEFAULT_CONFIG_CLASS):
            raise LoggingError('Logging configuration is not valid')
        name = name if name else __name__.split('.')[0]
        self.logger = self._create_logger(name, self.config)
        app.state.PLUGIN_LOGGER = self

    async def init(self) -> None:
        self.logger.info('Logging plugin is ON')

    async def terminate(self) -> None:
        self.logger.info('Logging plugin is OFF')
        self.config = None
        self.logger = None

    async def health(self) -> typing.Dict:
        return dict(level=self.logger.level, style=self.config.logging_style)


log_plugin = LoggingPlugin()


async def log_adapter(
    logger: typing.Union[logging.Logger, LoggerAdapter],
    extra: typing.Dict=None
) -> LoggerAdapter:
    if extra is None:
        extra = {}
    if isinstance(logger, logging.Logger):
        return LoggerAdapter(logger, extra)
    else:
        _extra = dict(**logger.extra)
        _extra.update(**extra)
        return LoggerAdapter(logger.logger, _extra)


async def depends_logging(
    conn: starlette.requests.HTTPConnection
) -> logging.Logger:
    return await conn.app.state.PLUGIN_LOGGER()


TLoggerPlugin = Annotated[logging.Logger, fastapi.Depends(depends_logging)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# tests.test_logger

from __future__ import absolute_import

import json
import logging

import fastapi
import pytest

import fastapi_plugins


pytestmark = [pytest.mark.anyio, pytest.mark.logger]


@pytest.fixture(params=None)
async def logapp(request):
    app = fastapi_plugins.register_middleware(fastapi.FastAPI())
    await fastapi_plugins.log_plugin.init_app(
        app=app,
        config=request.param or fastapi_plugins.LoggingSettings(),
        name=request.node.name
    )
    await fastapi_plugins.log_plugin.init()
    yield app
    await fastapi_plugins.log_plugin.terminate()


@pytest.fixture
async def logapp_name(request):
    return request.node.name


@pytest.mark.parametrize(
    'logapp, level',
    [
        pytest.param(None, logging.WARNING),
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.CRITICAL
            ),
            logging.CRITICAL
        ),
    ],
    indirect=['logapp']
)
async def test_basic(logapp, level, logapp_name):
    logger = await fastapi_plugins.log_plugin()
    assert logapp_name == logger.name
    assert level == logger.level


@pytest.mark.parametrize(
    'logapp, level, result',
    [
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logtxt,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            ['Hello', 'World', 'Echo']
        ),
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logjson,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            [
                {"message": "Hello", "level": "DEBUG"},
                {"message": "World", "planet": "earth", "level": "INFO"},
                {"message": "Echo", "planet": "earth", "satellite": ["moon"], "level": "WARNING"}   # noqa E501
            ]
        ),
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logfmt,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            [
                'at=DEBUG msg="Hello" process=MainProcess',
                'at=INFO msg="World" process=MainProcess planet="earth"',
                'at=WARNING msg="Echo" process=MainProcess planet="earth" satellite="[\'moon\']"'   # noqa E501
            ]
        ),
    ],
    indirect=['logapp']
)
async def test_format(logapp, level, result, logapp_name):
    def _preproc(_results):
        def _inner(__results):
            for r in __results:
                try:
                    rr = json.loads(r)
                    rr.pop('timestamp')
                    rr.pop('name')
                    yield rr
                except json.decoder.JSONDecodeError:
                    yield r
        return list(_inner(_results))

    logger = await fastapi_plugins.log_plugin()
    assert logapp_name == logger.name
    assert level == logger.level
    #
    logger.debug('Hello')
    logger.info('World', extra=dict(planet='earth'))
    logger.warning('Echo', extra=dict(planet='earth', satellite=['moon']))
    h = logger.handlers[0]
    assert result == _preproc([r for r in h.mqueue.queue][1:])


@pytest.mark.parametrize(
    'logapp, level, result',
    [
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logtxt,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            ['Hello', 'World', 'Echo']
        ),
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logjson,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            [
                {"message": "Hello", "level": "DEBUG"},
                {"message": "World", "planet": "earth", "level": "INFO"},
                {"message": "Echo", "planet": "earth", "satellite": ["moon"], "level": "WARNING"}   # noqa E501
            ]
        ),
        pytest.param(
            fastapi_plugins.LoggingSettings(
                logging_level=logging.DEBUG,
                logging_style=fastapi_plugins.LoggingStyle.logfmt,
                logging_handler=fastapi_plugins.LoggingHandlerType.loglist
            ),
            logging.DEBUG,
            [
                'at=DEBUG msg="Hello" process=MainProcess',
                'at=INFO msg="World" process=MainProcess planet="earth"',
                'at=WARNING msg="Echo" process=MainProcess planet="earth" satellite="[\'moon\']"'   # noqa E501
            ]
        ),
    ],
    indirect=['logapp']
)
async def test_format_adapter(logapp, level, result, logapp_name):
    def _preproc(_results):
        def _inner(__results):
            for r in __results:
                try:
                    rr = json.loads(r)
                    rr.pop('timestamp')
                    rr.pop('name')
                    yield rr
                except json.decoder.JSONDecodeError:
                    yield r
        return list(_inner(_results))

    logger = await fastapi_plugins.log_plugin()
    assert logapp_name == logger.name
    assert level == logger.level
    #
    logger.debug('Hello')
    logger = await fastapi_plugins.log_adapter(logger, extra=dict(planet='earth'))      # noqa E501
    logger.info('World')
    logger = await fastapi_plugins.log_adapter(logger, extra=dict(satellite=['moon']))  # noqa E501
    logger.warning('Echo')
    h = logger.logger.handlers[0]
    assert result == _preproc([r for r in h.mqueue.queue][1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.middleware

from __future__ import absolute_import

import typing

import fastapi
import starlette.middleware.cors


def register_middleware(
        app: fastapi.FastAPI,
        middleware: typing.List[typing.Tuple[type, typing.Any]]=None
) -> fastapi.FastAPI:
    if not middleware:
        middleware = [
            (
                starlette.middleware.cors.CORSMiddleware,
                dict(
                    allow_origins=["*"],
                    allow_credentials=True,
                    allow_methods=["*"],
                    allow_headers=["*"]
                )
            )
        ]
    for mw_klass, options in middleware:
        app.add_middleware(mw_klass, **options)
    return app



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.utils

from __future__ import absolute_import

import sys

if sys.version_info >= (3, 9):
    from typing import Annotated                                    # noqa
else:
    from typing_extensions import Annotated                         # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# setup

import codecs
import importlib.util
import os
import sys

from setuptools import setup
from setuptools import find_packages

__author__ = 'madkote <madkote(at)bluewin.ch>'
__copyright__ = 'Copyright 2023, madkote'


if sys.version_info < (3, 6, 0):
    raise RuntimeError("Python 3.6+ required")

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("CHANGES.md", "r") as fh:
    changes_description = fh.read()


def get_version(package_name):
    try:
        version_file = os.path.join(
            os.path.dirname(__file__),
            package_name,
            'version.py'
        )
        spec = importlib.util.spec_from_file_location(
            '%s.version' % package_name,
            version_file
        )
        version_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(version_module)
        package_version = version_module.__version__
    except Exception as e:
        raise ValueError(
            'can not determine "%s" version: %s :: %s' % (
                package_name, type(e), e
            )
        )
    else:
        return package_version


def load_requirements(filename):
    def _is_req(s):
        return s and not s.startswith('-r ') and not s.startswith('git+http') and not s.startswith('#')  # noqa E501

    def _read():
        with codecs.open(filename, 'r', encoding='utf-8-sig') as fh:
            for line in fh:
                line = line.strip()
                if _is_req(line):
                    yield line
    return sorted(list(_read()))


def package_files(directory=None):
    paths = []
    if directory:
        for (path, _, filenames) in os.walk(directory):
            for filename in filenames:
                paths.append(os.path.join('..', path, filename))
    return paths


NAME = 'fastapi-plugins'
NAME_PACKAGE = NAME.replace('-', '_')
VERSION = get_version(NAME_PACKAGE)
DESCRIPTION = 'Plugins for FastAPI framework'
URL = 'https://github.com/madkote/%s' % NAME

REQUIRES_INSTALL = [
    'fastapi>=0.74.0',
    'pydantic>=1.0.0,<2.0.0',
    'tenacity>=8.0.0',
    #
    'python-json-logger>=2.0.0',
    'redis[hiredis]>=4.3.0',
    'aiojobs>=1.0.0'
]
REQUIRES_FAKEREDIS = ['fakeredis[lua]>=1.8.0']
REQUIRES_MEMCACHED = ['aiomcache>=0.7.0']
REQUIRES_TESTS = [
    'bandit',
    'docker-compose',
    'flake8',
    'pytest',
    'pytest-asyncio',
    'pytest-cov',
    'tox',
    'twine',
    #
    'fastapi[all]',
]

REQUIRES_EXTRA = {
    'all': REQUIRES_MEMCACHED,
    'fakeredis': REQUIRES_FAKEREDIS,
    'memcached': REQUIRES_MEMCACHED,
    'dev': REQUIRES_MEMCACHED + REQUIRES_FAKEREDIS + REQUIRES_TESTS,
}

PACKAGES = find_packages(exclude=('scripts', 'tests'))
PACKAGE_DATA = {'': []}


# =============================================================================
# SETUP
# =============================================================================
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author='madkote',
    author_email=__author__.replace('(at)', '@'),
    url=URL,
    download_url=URL + '/archive/{}.tar.gz'.format(VERSION),
    license='MIT License',
    keywords=[
        'async', 'redis', 'aioredis', 'json', 'asyncio', 'plugin', 'fastapi',
        'aiojobs', 'scheduler', 'starlette', 'memcached', 'aiomcache'
    ],
    install_requires=REQUIRES_INSTALL,
    tests_require=REQUIRES_TESTS,
    extras_require=REQUIRES_EXTRA,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    python_requires='>=3.6.0',
    include_package_data=True,
    long_description='\n\n'.join((long_description, changes_description)),
    long_description_content_type='text/markdown',
    platforms=['any'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Logging',
        'Framework :: AsyncIO',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.memcached

from __future__ import absolute_import

import typing

try:
    import aiomcache
except ImportError:
    raise RuntimeError('aiomcache is not installed')

import fastapi
import pydantic
import starlette.requests
import tenacity

from .plugin import PluginError
from .plugin import PluginSettings
from .plugin import Plugin

from .control import ControlHealthMixin
from .utils import Annotated

__all__ = [
    'MemcachedError', 'MemcachedSettings', 'MemcachedClient',
    'MemcachedPlugin', 'memcached_plugin', 'depends_memcached',
    'TMemcachedPlugin'
]


class MemcachedError(PluginError):
    pass


class MemcachedSettings(PluginSettings):
    memcached_host: str = 'localhost'
    memcached_port: int = 11211
    memcached_pool_size: int = 10
    memcached_pool_minsize: int = 1
    #
    # TODO: xxx - should be shared across caches
    memcached_prestart_tries: int = 60 * 5  # 5 min
    memcached_prestart_wait: int = 1        # 1 second


class MemcachedClient(aiomcache.Client):
    async def ping(self) -> bytes:
        return await self.version()


class MemcachedPlugin(Plugin, ControlHealthMixin):
    DEFAULT_CONFIG_CLASS = MemcachedSettings

    def _on_init(self) -> None:
        self.memcached: MemcachedClient = None

    async def _on_call(self) -> MemcachedClient:
        if self.memcached is None:
            raise MemcachedError('Memcached is not initialized')
        return self.memcached

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        if self.config is None:
            raise MemcachedError('Memcached configuration is not initialized')
        elif not isinstance(self.config, self.DEFAULT_CONFIG_CLASS):
            raise MemcachedError('Memcached configuration is not valid')
        app.state.MEMCACHED = self

    async def init(self):
        if self.memcached is not None:
            raise MemcachedError('Memcached is already initialized')
        self.memcached = MemcachedClient(
            host=self.config.memcached_host,
            port=self.config.memcached_port,
            pool_size=2,
            pool_minsize=None
        )

        @tenacity.retry(
            stop=tenacity.stop_after_attempt(
                self.config.memcached_prestart_tries
            ),
            wait=tenacity.wait_fixed(
                self.config.memcached_prestart_wait
            ),
        )
        async def _init_memcached():
            await self.memcached.version()

        try:
            await _init_memcached()
        except Exception as e:
            raise MemcachedError(
                'Memcached initialization failed :: %s :: %s' % (type(e), e)
            )

    async def terminate(self):
        self.config = None
        if self.memcached is not None:
            await self.memcached.flush_all()
            await self.memcached.close()
            self.memcached = None

    async def health(self) -> typing.Dict:
        return dict(
            host=self.config.memcached_host,
            port=self.config.memcached_port,
            version=(await self.memcached.ping()).decode()
        )


memcached_plugin = MemcachedPlugin()


async def depends_memcached(
    conn: starlette.requests.HTTPConnection
) -> MemcachedClient:
    return await conn.app.state.MEMCACHED()


TMemcachedPlugin = Annotated[MemcachedClient, fastapi.Depends(depends_memcached)]   # noqa E501



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.plugin

from __future__ import absolute_import

import abc
import typing

import fastapi
import pydantic


class PluginError(Exception):
    pass


class PluginSettings(pydantic.BaseSettings):
    class Config:
        env_prefix = ''
        use_enum_values = True


class Plugin:
    DEFAULT_CONFIG_CLASS: pydantic.BaseSettings = None

    def __init__(
            self,
            app: fastapi.FastAPI=None,
            config: pydantic.BaseSettings=None
    ):
        self._on_init()
        if app and config:
            self.init_app(app, config)

    def __call__(self) -> typing.Any:
        return self._on_call()

    def _on_init(self) -> None:
        pass

#     def is_initialized(self) -> bool:
#         raise NotImplementedError('implement is_initialied()')

    @abc.abstractmethod
    async def _on_call(self) -> typing.Any:
        raise NotImplementedError('implement _on_call()')

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,
            *args,
            **kwargs
    ) -> None:
        pass

    async def init(self):
        pass

    async def terminate(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.control
#
# Health is inspired by
# "https://dzone.com/articles/an-overview-of-health-check-patterns"
#

from __future__ import absolute_import

import abc
import asyncio
import pprint
import typing

import fastapi
import pydantic
import starlette.requests

from .plugin import PluginError
from .plugin import PluginSettings
from .plugin import Plugin
from .utils import Annotated

__all__ = [
    'ControlEnviron', 'ControlHealthCheck', 'ControlHealth',
    'ControlHealthError', 'ControlHeartBeat', 'ControlVersion',
    #
    'ControlError', 'ControlHealthMixin', 'ControlSettings', 'Controller',
    'ControlPlugin', 'control_plugin', 'depends_control', 'TControlPlugin'
]


DEFAULT_CONTROL_ROUTER_PREFIX = 'control'
DEFAULT_CONTROL_VERSION = '0.0.1'


class ControlError(PluginError):
    pass


class ControlBaseModel(pydantic.BaseModel):
    class Config:
        use_enum_values = True
        validate_all = True


class ControlEnviron(ControlBaseModel):
    environ: typing.Dict = pydantic.Field(
        ...,
        title='Environment',
        example=dict(var1='variable1', var2='variable2')
    )


class ControlHealthStatus(ControlBaseModel):
    status: bool = pydantic.Field(
        ...,
        title='Health status',
        example=True
    )


class ControlHealthCheck(ControlHealthStatus):
    name: str = pydantic.Field(
        ...,
        title='Health check name',
        min_length=1,
        example='Redis'
    )
    details: typing.Dict = pydantic.Field(
        ...,
        title='Health check details',
        example=dict(detail1='detail1', detail2='detail2')
    )


class ControlHealth(ControlHealthStatus):
    checks: typing.List[ControlHealthCheck] = pydantic.Field(
        ...,
        title='Health checks',
        example=[
            ControlHealthCheck(
                name='Redis',
                status=True,
                details=dict(
                    redis_type='redis',
                    redis_host='localhost',
                )
            ).dict()
        ]
    )


class ControlHealthError(ControlBaseModel):
    detail: ControlHealth = pydantic.Field(
        ...,
        title='Health error',
        example=ControlHealth(
            status=False,
            checks=[
                ControlHealthCheck(
                    name='Redis',
                    status=False,
                    details=dict(error='Some error')
                )
            ]
        )
    )


class ControlHeartBeat(ControlBaseModel):
    is_alive: bool = pydantic.Field(
        ...,
        title='Alive flag',
        example=True
    )


# class ControlInfo(ControlBaseModel):
#     status: str = pydantic.Field(
#         'API is up and running',
#         title='status',
#         min_length=1,
#         exampe='API is up and running'
#     )


class ControlVersion(ControlBaseModel):
    version: str = pydantic.Field(
        ...,
        title='Version',
        min_length=1,
        example='1.2.3'
    )


class ControlHealthMixin(object):
    @abc.abstractmethod
    async def health(self) -> typing.Dict:
        pass


class Controller(object):
    def __init__(
            self,
            router_prefix: str=DEFAULT_CONTROL_ROUTER_PREFIX,
            router_tag: str=DEFAULT_CONTROL_ROUTER_PREFIX,
            version: str=DEFAULT_CONTROL_VERSION,
            environ: typing.Dict=None,
            failfast: bool=True
    ):
        self.router_prefix = router_prefix
        self.router_tag = router_tag
        self.version = version
        self.environ = environ
        self.plugins: typing.List[ControlHealthMixin] = []
        self.failfast = failfast

    def patch_app(
            self,
            app: fastapi.FastAPI,
            enable_environ: bool=True,
            enable_health: bool=True,
            enable_heartbeat: bool=True,
            enable_version: bool=True
    ) -> None:
        #
        # register plugins
        for name, state in app.state._state.items():
            if isinstance(state, ControlHealthMixin):
                self.plugins.append((name, state))
        #
        # register endpoints
        if not (enable_environ or enable_health or enable_heartbeat or enable_version): # noqa E501
            return

        router_control = fastapi.APIRouter()

        if enable_version:
            @router_control.get(
                '/version',
                summary='Version',
                description='Get the version',
                response_model=ControlVersion
            )
            async def version_get() -> ControlVersion:
                return ControlVersion(version=await self.get_version())

        if enable_environ:
            @router_control.get(
                '/environ',
                summary='Environment',
                description='Get the environment',
                response_model=ControlEnviron
            )
            async def environ_get() -> ControlEnviron:
                return ControlEnviron(
                    environ=dict(**(await self.get_environ()))
                )

        if enable_heartbeat:
            @router_control.get(
                '/heartbeat',
                summary='Heart beat',
                description='Get the alive signal',
                response_model=ControlHeartBeat
            )
            async def heartbeat_get() -> ControlHeartBeat:
                return ControlHeartBeat(is_alive=await self.get_heart_beat())

        if enable_health:
            @router_control.get(
                '/health',
                summary='Health',
                description='Get the health',
                response_model=ControlHealth,
                responses={
                    starlette.status.HTTP_200_OK: dict(
                        description='UP and healthy',
                        model=ControlHealth
                    ),
                    starlette.status.HTTP_417_EXPECTATION_FAILED: dict(
                        description='NOT healthy',
                        model=ControlHealthError
                    )
                }
            )
            async def health_get() -> ControlHealth:
                health = await self.get_health()
                if health.status:
                    return health
                else:
                    raise fastapi.HTTPException(
                        status_code=starlette.status.HTTP_417_EXPECTATION_FAILED,   # noqa E501
                        detail=health.dict()
                    )

        #
        # register router
        app.include_router(
            router_control,
            prefix='' if not self.router_prefix else '/' + self.router_prefix,
            tags=[self.router_tag],
        )

    async def register_plugin(self, plugin: ControlHealthMixin):
        raise NotImplementedError

    async def get_environ(self) -> typing.Dict:
        return self.environ if self.environ is not None else {}

    async def get_health(self) -> ControlHealth:
        shared_obj = type('', (), {})()
        shared_obj.status = True

        # TODO: implement failfast -> wait()
        async def wrappit(name, hfunc) -> ControlHealthCheck:
            try:
                details = await hfunc or {}
                status = True
            except Exception as e:
                details = dict(error=str(e))
                status = False
                shared_obj.status = False
            finally:
                return ControlHealthCheck(
                    name=name,
                    status=status,
                    details=details
                )

        # TODO: perform all checks with this function asyncio.wait(fs)
        # TODO: perform without shared object
        results = await asyncio.gather(
            *[wrappit(name, plugin.health()) for name, plugin in self.plugins]
        )
        return ControlHealth(
            status=shared_obj.status,
            checks=results
        )

    async def get_heart_beat(self) -> bool:
        return True

    async def get_version(self) -> str:
        return self.version


class ControlSettings(PluginSettings):
    control_router_prefix: str = DEFAULT_CONTROL_ROUTER_PREFIX
    control_router_tag: str = DEFAULT_CONTROL_ROUTER_PREFIX
    control_enable_environ: bool = True
    control_enable_health: bool = True
    control_enable_heartbeat: bool = True
    control_enable_version: bool = True


class ControlPlugin(Plugin):
    DEFAULT_CONFIG_CLASS = ControlSettings

    def _on_init(self) -> None:
        self.controller: Controller = None

    async def _on_call(self) -> Controller:
        if self.controller is None:
            raise ControlError('Control is not initialized')
        return self.controller

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None,
            *,
            version: str=DEFAULT_CONTROL_VERSION,
            environ: typing.Dict=None
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        if self.config is None:
            raise ControlError('Control configuration is not initialized')
        elif not isinstance(self.config, self.DEFAULT_CONFIG_CLASS):
            raise ControlError('Control configuration is not valid')
        app.state.PLUGIN_CONTROL = self
        #
        # initialize here while `app` is available
        self.controller = Controller(
            router_prefix=self.config.control_router_prefix,
            router_tag=self.config.control_router_tag,
            version=version,
            environ=environ
        )
        self.controller.patch_app(
            app,
            enable_environ=self.config.control_enable_environ,
            enable_health=self.config.control_enable_health,
            enable_heartbeat=self.config.control_enable_heartbeat,
            enable_version=self.config.control_enable_version
        )

    async def init(self):
        if self.controller is None:
            raise ControlError('Control cannot be initialized')
        if self.config.control_enable_health:
            health = await self.controller.get_health()
            if not health.status:
                print()
                print('-' * 79)
                pprint.pprint(health.dict())
                print('-' * 79)
                print()
                raise ControlError('failed health control')

    async def terminate(self):
        self.config = None
        self.controller = None


control_plugin = ControlPlugin()


async def depends_control(
    conn: starlette.requests.HTTPConnection
) -> Controller:
    return await conn.app.state.PLUGIN_CONTROL()


TControlPlugin = Annotated[Controller, fastapi.Depends(depends_control)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins._redis

from __future__ import absolute_import

import enum
import typing

import fastapi
import pydantic
import redis.asyncio as aioredis
import redis.asyncio.sentinel as aioredis_sentinel
import starlette.requests
import tenacity

from .plugin import PluginError
from .plugin import PluginSettings
from .plugin import Plugin

from .control import ControlHealthMixin
from .utils import Annotated

__all__ = [
    'RedisError', 'RedisType', 'RedisSettings', 'RedisPlugin',
    'redis_plugin', 'depends_redis', 'TRedisPlugin'
]


class RedisError(PluginError):
    pass


@enum.unique
class RedisType(str, enum.Enum):
    redis = 'redis'
    sentinel = 'sentinel'
    fakeredis = 'fakeredis'
    # cluster = 'cluster'


class RedisSettings(PluginSettings):
    redis_type: RedisType = RedisType.redis
    #
    redis_url: str = None
    redis_host: str = 'localhost'
    redis_port: int = 6379
    redis_user: str = None
    redis_password: str = None
    redis_db: int = None
    # redis_connection_timeout: int = 2
    #
    # redis_pool_minsize: int = 1
    # redis_pool_maxsize: int = None
    redis_max_connections: int = None
    redis_decode_responses: bool = True
    #
    redis_ttl: int = 3600
    #
    # TODO: xxx the customer validator does not work
    # redis_sentinels: typing.List = None
    redis_sentinels: str = None
    redis_sentinel_master: str = 'mymaster'
    #
    # TODO: xxx - should be shared across caches
    redis_prestart_tries: int = 60 * 5  # 5 min
    redis_prestart_wait: int = 1        # 1 second

    def get_redis_address(self) -> str:
        if self.redis_url:
            return self.redis_url
        elif self.redis_db:
            return 'redis://%s:%s/%s' % (
                self.redis_host,
                self.redis_port,
                self.redis_db
            )
        else:
            return 'redis://%s:%s' % (self.redis_host, self.redis_port)

    # TODO: xxx the customer validator does not work
    def get_sentinels(self) -> typing.List:
        if self.redis_sentinels:
            try:
                return [
                    (
                        _conn.split(':')[0].strip(),
                        int(_conn.split(':')[1].strip())
                    )
                    for _conn in self.redis_sentinels.split(',')
                    if _conn.strip()
                ]
            except Exception as e:
                raise RuntimeError(
                    'bad sentinels string :: %s :: %s :: %s' % (
                        type(e), str(e), self.redis_sentinels
                    )
                )
        else:
            return []


class RedisPlugin(Plugin, ControlHealthMixin):
    DEFAULT_CONFIG_CLASS = RedisSettings

    def _on_init(self) -> None:
        self.redis: typing.Union[aioredis.Redis, aioredis_sentinel.Sentinel] = None # noqa E501

    async def _on_call(self) -> typing.Any:
        if self.redis is None:
            raise RedisError('Redis is not initialized')
        #
        if self.config.redis_type == RedisType.sentinel:
            conn = self.redis.master_for(self.config.redis_sentinel_master)
        elif self.config.redis_type == RedisType.redis:
            conn = self.redis
        elif self.config.redis_type == RedisType.fakeredis:
            conn = self.redis
        else:
            raise NotImplementedError(
                'Redis type %s is not implemented' % self.config.redis_type
            )
        #
        conn.TTL = self.config.redis_ttl
        return conn

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        if self.config is None:
            raise RedisError('Redis configuration is not initialized')
        elif not isinstance(self.config, self.DEFAULT_CONFIG_CLASS):
            raise RedisError('Redis configuration is not valid')
        app.state.REDIS = self

    async def init(self):
        if self.redis is not None:
            raise RedisError('Redis is already initialized')
        #
        opts = dict(
            db=self.config.redis_db,
            username=self.config.redis_user,
            password=self.config.redis_password,
            # minsize=self.config.redis_pool_minsize,
            # maxsize=self.config.redis_pool_maxsize,
            max_connections=self.config.redis_max_connections,
            decode_responses=self.config.redis_decode_responses,
        )
        #
        if self.config.redis_type == RedisType.redis:
            address = self.config.get_redis_address()
            method = aioredis.from_url
            # opts.update(dict(timeout=self.config.redis_connection_timeout))
        elif self.config.redis_type == RedisType.fakeredis:
            try:
                import fakeredis.aioredis
            except ImportError:
                raise RedisError(f'{self.config.redis_type} requires fakeredis to be installed')    # noqa E501
            else:
                address = self.config.get_redis_address()
                method = fakeredis.aioredis.FakeRedis.from_url
        elif self.config.redis_type == RedisType.sentinel:
            address = self.config.get_sentinels()
            method = aioredis_sentinel.Sentinel
        else:
            raise NotImplementedError(
                'Redis type %s is not implemented' % self.config.redis_type
            )
        #
        if not address:
            raise ValueError('Redis address is empty')

        @tenacity.retry(
            stop=tenacity.stop_after_attempt(self.config.redis_prestart_tries),
            wait=tenacity.wait_fixed(self.config.redis_prestart_wait),
        )
        async def _inner():
            return method(address, **opts)

        self.redis = await _inner()
        await self.ping()

    async def terminate(self):
        self.config = None
        if self.redis is not None:
            # del self.redis
            self.redis = None
            # self.redis.close()
            # await self.redis.wait_closed()
            # self.redis = None

    async def health(self) -> typing.Dict:
        return dict(
            redis_type=self.config.redis_type,
            redis_address=self.config.get_sentinels() if self.config.redis_type == RedisType.sentinel else self.config.get_redis_address(),  # noqa E501
            redis_pong=(await self.ping())
        )

    async def ping(self):
        if self.config.redis_type == RedisType.redis:
            return await self.redis.ping()
        elif self.config.redis_type == RedisType.fakeredis:
            return await self.redis.ping()
        elif self.config.redis_type == RedisType.sentinel:
            return await self.redis.master_for(self.config.redis_sentinel_master).ping()    # noqa E501
        else:
            raise NotImplementedError(
                'Redis type %s.ping() is not implemented' % self.config.redis_type          # noqa E501
            )


redis_plugin = RedisPlugin()


async def depends_redis(
    conn: starlette.requests.HTTPConnection
) -> aioredis.Redis:
    return await conn.app.state.REDIS()


TRedisPlugin = Annotated[typing.Any, fastapi.Depends(depends_redis)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# fastapi_plugins.scheduler

from __future__ import absolute_import

import typing

import aiojobs
import fastapi
import pydantic
import starlette.requests

from .plugin import PluginError
from .plugin import PluginSettings
from .plugin import Plugin

from .control import ControlHealthMixin
from .utils import Annotated
from .version import VERSION

__all__ = [
    'SchedulerError', 'SchedulerSettings', 'SchedulerPlugin',
    'scheduler_plugin', 'depends_scheduler', 'TSchedulerPlugin'
    # 'MadnessScheduler'
]
__author__ = 'madkote <madkote(at)bluewin.ch>'
__version__ = '.'.join(str(x) for x in VERSION)
__copyright__ = 'Copyright 2021, madkote'


class SchedulerError(PluginError):
    pass


# class MadnessScheduler(aiojobs.Scheduler):
#     def __init__(self, *args, **kwargs):
#         super(MadnessScheduler, self).__init__(*args, **kwargs)
#         self._job_map = {}
#
#     async def spawn(self, coro):
#         job = await aiojobs.Scheduler.spawn(self, coro)
#         self._job_map[job.id] = job
#         return job
#
#     def _done(self, job) -> None:
#         if job.id in self._job_map:
#             self._job_map.pop(job.id)
#         return aiojobs.Scheduler._done(self, job)
#
#     async def cancel_job(self, job_id: str) -> None:
#         if job_id in self._job_map:
#             await self._job_map[job_id].close()
#
#
# async def create_madness_scheduler(
#     *args,
#     close_timeout=0.1,
#     limit=100,
#     pending_limit=10000,
#     exception_handler=None
# ) -> MadnessScheduler:
#     if exception_handler is not None and not callable(exception_handler):
#         raise TypeError(
#             'A callable object or None is expected, got {!r}'.format(
#                 exception_handler
#             )
#         )
#     loop = asyncio.get_event_loop()
#     return MadnessScheduler(
#         loop=loop,
#         close_timeout=close_timeout,
#         limit=limit,
#         pending_limit=pending_limit,
#         exception_handler=exception_handler
#     )


# TODO: test settings (values)
# TODO: write unit tests
class SchedulerSettings(PluginSettings):
    aiojobs_close_timeout: float = 0.1
    aiojobs_limit: int = 100
    aiojobs_pending_limit: int = 10000
    # aiojobs_enable_cancel: bool = False


class SchedulerPlugin(Plugin, ControlHealthMixin):
    DEFAULT_CONFIG_CLASS = SchedulerSettings

    def _on_init(self) -> None:
        self.scheduler: aiojobs.Scheduler = None

    async def _on_call(self) -> aiojobs.Scheduler:
        if self.scheduler is None:
            raise SchedulerError('Scheduler is not initialized')
        return self.scheduler

    async def init_app(
            self,
            app: fastapi.FastAPI,
            config: pydantic.BaseSettings=None
    ) -> None:
        self.config = config or self.DEFAULT_CONFIG_CLASS()
        if self.config is None:
            raise SchedulerError('Scheduler configuration is not initialized')
        elif not isinstance(self.config, self.DEFAULT_CONFIG_CLASS):
            raise SchedulerError('Scheduler configuration is not valid')
        app.state.AIOJOBS_SCHEDULER = self

    async def init(self):
        if self.scheduler is not None:
            raise SchedulerError('Scheduler is already initialized')
        self.scheduler = aiojobs.Scheduler(
            close_timeout=self.config.aiojobs_close_timeout,
            limit=self.config.aiojobs_limit,
            pending_limit=self.config.aiojobs_pending_limit
        )

    async def terminate(self):
        self.config = None
        if self.scheduler is not None:
            await self.scheduler.close()
            self.scheduler = None

    async def health(self) -> typing.Dict:
        return dict(
            jobs=len(self.scheduler),
            active=self.scheduler.active_count,
            pending=self.scheduler.pending_count,
            limit=self.scheduler.limit,
            closed=self.scheduler.closed
        )


scheduler_plugin = SchedulerPlugin()


async def depends_scheduler(
    conn: starlette.requests.HTTPConnection
) -> aiojobs.Scheduler:
    return await conn.app.state.AIOJOBS_SCHEDULER()


TSchedulerPlugin = Annotated[aiojobs.Scheduler, fastapi.Depends(depends_scheduler)] # noqa E501
